# coding=utf-8

import getpass
import os
import pydoc

import pkg_resources
from typing import Callable
from . import coordinator


def extra(name: str, desc: str) -> Callable:
    """
    Decorator for slave channel's "additional features" interface.

    Args:
        name (str): A human readable name for the function.
        desc (str): A short description and usage of it. Use
            ``{function_name}`` in place of the function name
            in the description.

    Returns:
        The decorated method.
    """

    def attr_dec(f):
        f.__setattr__("extra_fn", True)
        f.__setattr__("name", name)
        f.__setattr__("desc", desc)
        return f

    return attr_dec


def get_base_path() -> str:
    """
    Get the base data path for EFB. This can be defined by the
    environment variable ``EFB_DATA_PATH``.
    
    If ``EFB_DATA_PATH`` is not defined, this gives
    ``~/.ehforwarderbot``.
    
    This method creates the queried path if not existing.
    
    Returns:
        str: The base path.
    """
    base_path = os.environ.get("EFB_DATA_PATH", None)
    if base_path:
        base_path = os.path.join(base_path, getpass.getuser(), "")
    else:
        base_path = os.path.expanduser("~/.ehforwarderbot/")
    os.makedirs(base_path, exist_ok=True)
    return base_path


def get_data_path(module_id: str):
    """
    Get the path for permanent storage of a module.
    
    This method creates the queried path if not existing.
    
    Args:
        module_id (str): Module ID

    Returns:
        str: The data path of selected channel.
    """
    profile = coordinator.profile
    base_path = get_base_path()
    data_path = os.path.join(base_path, 'profiles', profile, module_id, "")
    os.makedirs(data_path, exist_ok=True)
    return data_path


def get_config_path(module_id: str=None, ext: str= 'yaml') -> str:
    """
    Get path for configuration file. Defaulted to
    ``~/.ehforwarderbot/profiles/profile_name/channel_id/config.yaml``.
    
    This method creates the queried path if not existing. The config file will
    not be created, however.
    
    Args:
        module_id (str): Module ID.
        ext (Optional[Str]): Extension name of the config file.
            Defaulted to ``"yaml"``.

    Returns:
        str: The path to the configuration file.
    """
    base_path = get_base_path()
    profile = coordinator.profile
    if module_id:
        config_path = get_data_path(module_id)
    else:
        config_path = os.path.join(base_path, 'profiles', profile)
    os.makedirs(config_path, exist_ok=True)
    return os.path.join(config_path, "config.%s" % ext)


def get_custom_modules_path() -> str:
    """
    Get the path to custom channels

    Returns:
        str: The path.
    """
    channel_path = os.path.join(get_base_path(), "modules")
    os.makedirs(channel_path, exist_ok=True)
    return channel_path


def locate_module(module_id: str, module_type: str=None):
    """
    Locate module by module ID

    Args:
        module_id: Module ID
        module_type: Type of module, one of ``'master'``, ``'slave'`` and ``'middleware'``
    """

    entry_point = None

    if module_type:
        entry_point = 'ehforwarderbot.%s' % module_type

    module_id = module_id.split('#', 1)[0]

    if entry_point:
        for i in pkg_resources.iter_entry_points(entry_point):
            if i.name == module_id:
                return i.load()

    return pydoc.locate(module_id)




# coding=utf-8

"""
Coordinator among channels.

Attributes:
    profile (str): Name of current profile..
    mutex (threading.Lock): Global interaction thread lock.
    master (EFBChannel): The running master channel object.
    slaves (Dict[str, EFBChannel]): Dictionary of running slave channel object.
        Keys are the unique identifier of the channel.
    middlewares (List[EFBMiddleware]): List of middlewares
"""

import threading
from typing import List, Dict, Optional

from . import EFBMsg
from .channel import EFBChannel
from .constants import ChannelType
from .exceptions import EFBChannelNotFound
from .middleware import EFBMiddleware
from .status import EFBStatus

profile: str = "default"
"""Current running profile name"""

mutex: threading.Lock = threading.Lock()
"""Mutual exclusive lock for user interaction through CLI interface"""

master: EFBChannel = None
"""The instance of the master channel."""

slaves: Dict[str, EFBChannel] = dict()
"""Instances of slave channels. Keys are the channel IDs."""

middlewares: List[EFBMiddleware] = list()
"""Instances of middlewares. Sorted in the order of execution."""

master_thread: Optional[threading.Thread] = None
"""The thread running poll() of the master channel."""

slave_threads: Dict[str, threading.Thread] = dict()
"""Threads running poll() from slave channels. Keys are the channel IDs."""

translator = None
"""Internal GNU gettext translator."""


def add_channel(channel: EFBChannel):
    """
    Register the channel with the coordinator.

    Args:
        channel (EFBChannel): Channel to register
    """
    global master, slaves
    if isinstance(channel, EFBChannel):
        if channel.channel_type == ChannelType.Slave:
            slaves[channel.channel_id] = channel
        else:
            master = channel
    else:
        raise TypeError("Channel instance is expected")


def add_middleware(middleware: EFBMiddleware):
    """
    Register a middleware with the coordinator.

    Args:
        middleware (EFBMiddleware): Middleware to register
    """
    global middlewares
    if isinstance(middleware, EFBMiddleware):
        middlewares.append(middleware)
    else:
        raise TypeError("Middleware instance is expected")


def send_message(msg: EFBMsg) -> Optional[EFBMsg]:
    """
    Deliver a message to the destination channel.

    Args:
        msg (EFBMsg): The message

    Returns:
        The message sent by the destination channel,
        includes the updated message ID from there.
        Returns ``None`` if the message is not sent.
    """
    global middlewares, master, slaves
    if msg is None:
        return

    # Go through middlewares
    for i in middlewares:
        msg = i.process_message(msg)
        if msg is None:
            return

    msg.verify()

    if msg.deliver_to.channel_id == master.channel_id:
        return master.send_message(msg)
    elif msg.deliver_to.channel_id in slaves:
        return slaves[msg.deliver_to.channel_id].send_message(msg)
    else:
        raise EFBChannelNotFound(msg)


def send_status(status: EFBStatus):
    """
    Deliver a message to the destination channel.

    Args:
        status (EFBStatus): The status
    """
    global middlewares, master
    if status is None:
        return

    # Go through middlewares
    for i in middlewares:
        status = i.process_status(status)
        if status is None:
            return

    status.verify()

    status.destination_channel.send_status(status)



# coding=utf-8

import threading
import logging
import argparse
import sys
import atexit
import mimetypes
import pkg_resources
import gettext

from . import config, utils
from . import coordinator
from .__version__ import __version__
from .channel import EFBChannel
from .middleware import EFBMiddleware

# gettext.install('ehforwarderbot', 'locale')
coordinator.translator = gettext.translation('ehforwarderbot',
                                             pkg_resources.resource_filename('ehforwarderbot', 'locale'),
                                             fallback=True)

_ = coordinator.translator.gettext
ngettext = coordinator.translator.ngettext

if sys.version_info < (3, 6):
    raise Exception(_("Python 3.6 or higher is required. Your version is {version}.").format(version=sys.version))

description = _("EH Forwarder Bot is an extensible chat tunnel framework which allows "
                "users to contact people from other chat platforms, and ultimately "
                "remotely control their accounts in other platforms.")

parser = argparse.ArgumentParser(description=description,
                                 epilog="GitHub: https://github.com/blueset/ehForwarderBot")

parser.add_argument("-v", '--verbose', action='store_true',
                    help=_("Enable verbose mode. (Level: Debug)"))
parser.add_argument("-V", "--version", action='store_true',
                    help=_("Show version numbers and exit."))
parser.add_argument("-p", "--profile",
                    help=_("Choose a profile to start with."),
                    default="default")


def stop_gracefully():
    logger = logging.getLogger(__name__)
    if isinstance(coordinator.master, EFBChannel):
        coordinator.master.stop_polling()
        logger.debug("Stop signal sent to master: %s" % coordinator.master.channel_name)
    for i in coordinator.slaves:
        if isinstance(coordinator.slaves[i], EFBChannel):
            coordinator.slaves[i].stop_polling()
            logger.debug("Stop signal sent to slave: %s" % coordinator.slaves[i].channel_name)
    if coordinator.master_thread and coordinator.master_thread.is_alive():
        coordinator.master_thread.join()
    for i in coordinator.slave_threads.values():
        if i.is_alive():
            i.join()


def init():
    """
    Initialize all channels.
    """

    logger = logging.getLogger(__name__)

    # Initialize mimetypes library
    mimetypes.init([pkg_resources.resource_filename('ehforwarderbot', 'mimetypes')])

    # Initialize all channels
    # (Load libraries and modules and init them with Queue `q`)

    conf = config.load_config()

    for i in conf['slave_channels']:
        logger.log(99, "\x1b[0;37;46m %s \x1b[0m", _("Initializing slave {}...").format(i))

        cls = utils.locate_module(i, 'slave')
        instance_id = i.split('#', 1)[1:]
        instance_id = (instance_id and instance_id[0]) or None
        coordinator.add_channel(cls(instance_id=instance_id))

        logger.log(99, "\x1b[0;37;42m %s \x1b[0m",
                   _("Slave channel {name} ({id}) # {instance_id} is initialized.")
                   .format(name=cls.channel_name, id=cls.channel_id,
                           instance_id=instance_id))

    logger.log(99, "\x1b[0;37;46m %s \x1b[0m",
               _("Initializing master {}...").format(conf['master_channel']))
    instance_id = conf['master_channel'].split('#', 1)[1:]
    instance_id = (instance_id and instance_id[0]) or None
    coordinator.add_channel(utils.locate_module(conf['master_channel'], 'master')
                            (instance_id=instance_id))
    logger.log(99, "\x1b[0;37;42m %s \x1b[0m",
               _("Master channel {name} ({id}) # {instance_id} is initialized.")
               .format(name=coordinator.master.channel_name,
                       id=coordinator.master.channel_id,
                       instance_id=instance_id))

    logger.log(99, "\x1b[1;37;42m %s \x1b[0m", _("All channels initialized."))
    for i in conf['middlewares']:
        logger.log(99, "\x1b[0;37;46m %s \x1b[0m", _("Initializing middleware {}...").format(i))
        cls = utils.locate_module(i, 'middleware')

        instance_id = i.split('#', 1)[1:]
        instance_id = (instance_id and instance_id[0]) or None
        coordinator.add_middleware(cls(instance_id=instance_id))
        logger.log(99, "\x1b[0;37;42m %s \x1b[0m",
                   _("Middleware {name} ({id}) # {instance_id} is initialized.")
                   .format(name=cls.middleware_name, id=cls.middleware_id,
                           instance_id=instance_id))

    logger.log(99, "\x1b[1;37;42m %s \x1b[0m", _("All middlewares are initialized."))

    coordinator.master_thread = threading.Thread(target=coordinator.master.poll)
    coordinator.slave_threads = {key: threading.Thread(target=coordinator.slaves[key].poll)
                                 for key in coordinator.slaves}


def poll():
    """
    Start threads for polling
    """
    coordinator.master_thread.start()
    for i in coordinator.slave_threads:
        coordinator.slave_threads[i].start()


def main():
    args = parser.parse_args()

    if getattr(args, "version", None):
        versions = _("EH Forwarder Bot\n"
                     "Version: {version}\n"
                     "Python version:\n"
                     "{py_version}").format(version=__version__, py_version=sys.version)
        try:
            conf = config.load_config()
            # Master channel
            master_channel: EFBChannel = utils.locate_module(conf['master_channel'], 'master')
            instance_id = conf['master_channel'].split('#', 1)[1:]
            instance_id = (instance_id and instance_id[0]) or _("Default profile")
            versions += "\n\n" + _("Master channel:") + "\n    " + _("{name} ({id}) {version} # {instance_id}") \
                .format(name=master_channel.channel_name,
                        id=master_channel.channel_id,
                        version=master_channel.__version__,
                        instance_id=instance_id)
            versions += "\n\n" + ngettext("Slave channel:", "Slave channels:", len(conf['slave_channels']))
            for i in conf['slave_channels']:
                instance_id = i.split('#', 1)[1:]
                instance_id = (instance_id and instance_id[0]) or _("Default profile")
                slave_channel: EFBChannel = utils.locate_module(i, 'slave')
                versions += "\n    " + _("{name} ({id}) {version}") \
                            .format(name=slave_channel.channel_name,
                                    id=slave_channel.channel_id,
                                    version=slave_channel.__version__,
                                    instance_id=instance_id)
            versions += "\n\n" + ngettext("Middleware:", "Middlewares:", len(conf['middlewares']))
            if conf['middlewares']:
                for i in conf['middlewares']:
                    instance_id = i.split('#', 1)[1:]
                    instance_id = (instance_id and instance_id[0]) or _("Default profile")
                    middleware: EFBMiddleware = utils.locate_module(i, 'middleware')
                    versions += "\n    " + _("{name} ({id}) {version} # {instance_id}") \
                                .format(name=middleware.middleware_name,
                                        id=middleware.middleware_name,
                                        version=middleware.__version__,
                                        instance_id=instance_id)
            else:
                versions += "\n    " + _("No middleware is enabled.")
        finally:
            print(versions)
    else:
        if getattr(args, "verbose", None):
            level = logging.DEBUG
        else:
            level = logging.ERROR
        logging.basicConfig(format='%(asctime)s [%(levelname)s]: %(name)s (%(module)s.%(funcName)s; '
                                   '%(filename)s:%(lineno)d) \n    %(message)s', level=level)
        logging.root.setLevel(level)

        # signal.signal(signal.SIGINT, stop_gracefully)
        # signal.signal(signal.SIGTERM, stop_gracefully)
        atexit.register(stop_gracefully)

        if args.profile:
            coordinator.profile = str(args.profile)

        init()
        poll()


if __name__ == '__main__':
    main()



# coding=utf-8

from abc import abstractmethod, ABC
from typing import Tuple, Optional

from . import EFBChannel, EFBMsg

__all__ = ["EFBStatus", "EFBChatUpdates", "EFBMemberUpdates", "EFBMessageRemoval"]


class EFBStatus(ABC):
    """
    Abstract class of a status

    Attributes:
        destination_channel (:obj:`.EFBChannel`): The
            channel that this status is sent to, usually
            the master channel.
    """
    @abstractmethod
    def __init__(self):
        self.destination_channel: 'EFBChannel' = None
        raise NotImplementedError()

    @abstractmethod
    def verify(self):
        raise NotImplementedError()


class EFBChatUpdates(EFBStatus):
    """EFBChatUpdates(channel: EFBChannel, new_chats: Optional[Tuple[str]]=tuple(), removed_chats: Optional[Tuple[str]]=tuple(), modified_chats: Optional[Tuple[str]]=tuple())

    Inform the master channel on updates of slave chats.

    Attributes:
        channel (:obj:`.EFBChannel`): Slave channel that issues the update
        new_chats (Optional[Tuple[str]]): Unique ID of new chats
        removed_chats (Optional[Tuple[str]]): Unique ID of removed chats
        modified_chats (Optional[Tuple[str]]): Unique ID of modified chats
    """
    def __init__(self, channel: 'EFBChannel', new_chats: Optional[Tuple[str]]=tuple(),
                 removed_chats: Optional[Tuple[str]]=tuple(), modified_chats: Optional[Tuple[str]]=tuple()):
        """__init__(channel: EFBChannel, new_chats: Optional[Tuple[str]]=tuple(), removed_chats: Optional[Tuple[str]]=tuple(), modified_chats: Optional[Tuple[str]]=tuple())

        Args:
            channel (:obj:`.EFBChannel`): Slave channel that issues the update
            new_chats (Optional[Tuple[str]]): Unique ID of new chats
            removed_chats (Optional[Tuple[str]]): Unique ID of removed chats
            modified_chats (Optional[Tuple[str]]): Unique ID of modified chats
        """
        self.channel: 'EFBChannel' = channel
        self.new_chats: Tuple[str] = new_chats
        self.removed_chats: Tuple[str] = removed_chats
        self.modified_chats: Tuple[str] = modified_chats
        self.destination_channel: 'EFBChannel' = coordinator.master

    def __str__(self):
        return "<EFBChatUpdates @ {s.channel.channel_name}; New: {s.new_chats}; " \
               "Removed: {s.removed_chats}; Modified: {s.modified_chats}>".format(s=self)

    def verify(self):
        if self.channel is None or not isinstance(self.channel, EFBChannel):
            raise ValueError("Channel is not valid.")


class EFBMemberUpdates(EFBStatus):
    """EFBMemberUpdates(channel: EFBChannel, chat_id: str, new_members: Optional[Tuple[str]]=tuple(), removed_members: Optional[Tuple[str]]=tuple(), modified_members: Optional[Tuple[str]]=tuple())

    Inform the master channel on updates of members in a slave chat.

    Attributes:
        channel (:obj:`.EFBChannel`): Slave channel that issues the update
        chat_id (str): Unique ID of the chat.
        new_members (Optional[Tuple[str]]): Unique ID of new members
        removed_members (Optional[Tuple[str]]): Unique ID of removed members
        modified_members (Optional[Tuple[str]]): Unique ID of modified members
    """

    def __init__(self, channel: 'EFBChannel', chat_id: str,
                 new_members: Optional[Tuple[str]]=tuple(), removed_members: Optional[Tuple[str]]=tuple(),
                 modified_members: Optional[Tuple[str]]=tuple()):
        """__init__(channel: EFBChannel, chat_id: str, new_members: Optional[Tuple[str]]=tuple(), removed_members: Optional[Tuple[str]]=tuple(), modified_members: Optional[Tuple[str]]=tuple())

        Args:
            channel (:obj:`.EFBChannel`): Slave channel that issues the update
            chat_id (str): Unique ID of the chat.
            new_members (Optional[Tuple[str]]): Unique ID of new members
            removed_members (Optional[Tuple[str]]): Unique ID of removed members
            modified_members (Optional[Tuple[str]]): Unique ID of modified members
        """
        self.channel: 'EFBChannel' = channel
        self.chat_id: str = chat_id
        self.new_members: Tuple[str] = new_members
        self.removed_members: Tuple[str] = removed_members
        self.modified_members: Tuple[str] = modified_members
        self.destination_channel: 'EFBChannel' = coordinator.master

    def __str__(self):
        return "<EFBMemberUpdates: {s.chat_id} @ {s.channel.channel_name}; New: {s.new_chats}; " \
               "Removed: {s.removed_chats}; Modified: {s.modified_chats}>".format(s=self)

    def verify(self):
        if self.channel is None or not isinstance(self.channel, EFBChannel):
            raise ValueError("Channel is not valid.")


class EFBMessageRemoval(EFBStatus):
    """EFBMessageRemoval(source_channel: EFBChannel, destination_channel: EFBChannel, message: EFBMsg)

    Inform a channel to remove a certain message.

    This is usually known as "delete from everyone", "delete from recipient",
    "recall a message", or "revoke a message" as well, depends on the IM.

    Some channels may not support removal of messages, and raises a
    :obj:`.exceptions.EFBOperationNotSupported` exception.

    Attributes:
        source_channel (:obj:`.EFBChannel`): Channel issued the status
        destination_channel (:obj:`.EFBChannel`): Channel the status is issued to
        message (:obj:`.EFBMsg`): Message to remove.
            This may not be a complete :obj:`.EFBMsg` object.
    """

    def __init__(self, source_channel: 'EFBChannel', destination_channel: 'EFBChannel', message: 'EFBMsg'):
        """__init__(source_channel: EFBChannel, destination_channel: EFBChannel, message: EFBMsg)

        Create a message removal status

        Try to provided as much as you can, if not, provide a minimum information
        in the channel:

        * Slave channel ID and chat ID (:attr:`message.chat.channel_id <.EFBChat.channel_id>`
          and :attr:`message.chat.chat_uid <.EFBChat.chat_uid>`)
        * Message unique ID from the slave channel (:attr:`message.uid <.EFBMsg.uid>`)

        Args:
            source_channel (:obj:`.EFBChannel`): Channel issued the status
            destination_channel (:obj:`.EFBChannel`): Channel the status is issued to
            message (:obj:`.EFBMsg`): Message to remove.
        """
        self.source_channel: 'EFBChannel' = source_channel
        self.destination_channel: 'EFBChannel' = destination_channel
        self.message: 'EFBMsg' = message

    def __str__(self):
        return "<EFBMessageRemoval: {s.message}; {s.source_channel.channel_name} " \
               "-> {s.destination_channel.channel_name}>".format(s=self)

    def verify(self):
        if self.source_channel is None or not isinstance(self.source_channel, EFBChannel):
            raise ValueError("Source channel is not valid.")
        if self.destination_channel is None or not isinstance(self.destination_channel, EFBChannel):
            raise ValueError("Destination channel is not valid.")
        if self.message is None or not isinstance(self.message, EFBMsg):
            raise ValueError("Message channel is not valid.")
        if not self.message.chat.channel_id or not self.message.chat.chat_uid or not self.message.uid:
            raise ValueError("Message does not contain the minimum information required")



# coding=utf-8

from enum import Enum


class MsgType(Enum):
    Text = "Text"
    """Text message"""

    Image = "Image"
    """Image (picture) message"""

    Audio = "Audio"
    """Audio message. Including music and voice message."""

    File = "File"
    """Messages sent as a file."""

    Location = "Location"
    """Location message."""

    Video = "Video"
    """Video message"""

    Link = "Link"
    """
    Message that is mainly one specific link, or a
    text message with one link preview.
    """

    Sticker = "Sticker"
    """
    Pictures sent with few text caption, usually a
    transparent background, and a limited number
    of options that is usually not from the user's
    photo gallery.
    """

    Status = "Status"
    """
    Status from a user in a chat, usually typing and
    uploading.
    """

    Unsupported = "Unsupported"
    """
    Any type of message that is not listed above.
    A text representation is required.
    """


class ChatType(Enum):
    User = "User"
    Group = "Group"
    System = "System"


class TargetType(Enum):
    Member = "Member"
    Message = "Message"
    Substitution = "Substitution"


class ChannelType(Enum):
    Master = "Master"
    Slave = "Slave"



# coding=utf-8

import copy
from typing import List, Dict, Any, Optional

from .channel import EFBChannel
from .constants import ChatType


__all__ = ['EFBChat']


class EFBChat:
    """
    EFB Chat object. This is used to represent a chat or a group member.

    Attributes:
        channel_id (str): Unique ID of the channel.
        channel_emoji (str): Emoji of the channel.
        channel_name (str): Name of the channel.
        chat_name (str): Name of the chat.
        chat_alias (str): Alternative name of the chat, usually set by user.
        chat_type (:obj:`.ChatType`): Type of the chat.
        chat_uid (str): Unique ID of the chat. This should be unique within the channel.
        is_chat (bool): Indicate if this object represents a chat. Defaulted to ``True``.
            This should be set to ``False`` when used on a group member.
        group (:obj:`.EFBChat` or None): The parent chat of the member. Only
            available to chat member objects. Defaulted to ``None``.
        members (list of :obj:`.EFBChat`): Provide a list of members
            in the group. Defaulted to an empty ``list``. You may want to extend this
            object and implement a ``@property`` method set for loading members on
            demand.
        vendor_specific (Dict[str, Any]): Any vendor specific attributes.
        is_self (bool): Indicate if this chat represents the user.
        is_system (bool): Indicate if this chat represents a system chat/member.
    """

    SELF_ID = "__self__"
    SYSTEM_ID = "__system__"

    def __init__(self, channel: Optional[EFBChannel]=None):
        """
        Args:
            channel (Optional[:obj:`.EFBChannel`]):
                Provide the channel object to fill :attr:`channel_name`,
                :attr:`channel_emoji`, and :attr:`channel_id` automatically.
        """
        self.channel_name: str = None
        self.channel_emoji: str = None
        self.channel_id: str = None
        if isinstance(channel, EFBChannel):
            self.channel_name: str = channel.channel_name
            self.channel_emoji: str = channel.channel_emoji
            self.channel_id: str = channel.channel_id

        self.chat_name: str = None
        self.chat_type: ChatType = None
        self.chat_alias: str = None
        self.chat_uid: str = None
        self.is_chat: bool = True
        self.members: List[EFBChat] = []
        self.group: Optional[EFBChat] = None
        self.vendor_specific: Dict[str, Any] = dict()

    def self(self) -> 'EFBChat':
        """
        Set the chat as yourself.
        In this context, "yourself" means the user behind the master channel.
        Every channel should relate this to the corresponding target.

        Returns:
            EFBChat: This object.
        """
        self.chat_name = "You"
        self.chat_alias = None
        self.chat_uid = EFBChat.SELF_ID
        self.chat_type = ChatType.User
        return self

    def system(self) -> 'EFBChat':
        """
        Set the chat as a system chat.
        Only set for channel-level and group-level system chats.

        Returns:
            EFBChat: This object.
        """
        self.chat_name = "System"
        self.chat_alias = None
        self.chat_uid = EFBChat.SYSTEM_ID
        self.chat_type = ChatType.User
        return self

    @property
    def is_self(self) -> bool:
        """If this chat represents the user"""
        return self.chat_uid == EFBChat.SELF_ID

    @property
    def is_system(self) -> bool:
        """If this chat is a system chat"""
        return self.chat_uid == EFBChat.SYSTEM_ID

    def copy(self) -> 'EFBChat':
        return copy.copy(self)

    def verify(self):
        """
        Verify the completeness of the data.

        Raises:
            ValueError: When this chat is invalid.
        """
        if any(i is None for i in (self.chat_uid, self.channel_id)):
            print(self.chat_uid, self.channel_id)
            raise ValueError("Chat data is incomplete.")
        if not isinstance(self.chat_type, ChatType):
            raise ValueError("Invalid chat type.")
        if self.chat_type == ChatType.Group:
            if any(not isinstance(i, EFBChat) or not i.chat_type == ChatType.User for i in self.members):
                raise ValueError("The group has an invalid member.")
        if self.group is not None and (not isinstance(self.group, EFBChat) or
                                       not self.group.chat_type == ChatType.Group):
            raise ValueError("The member is in an invalid group.")

    def __eq__(self, other):
        return self.channel_id == other.channel_id and self.chat_uid == other.chat_uid

    def __str__(self):
        return "<EFBChat: {c.chat_name} ({alias}{c.chat_uid}) @ {c.channel_name}>" \
            .format(c=self, alias=self.chat_alias + ", " if self.chat_alias else "")



# coding=utf-8


class EFBException(Exception):
    """A general class to indicate that the exception is from EFB framework."""
    pass


class EFBChatNotFound(EFBException):
    """
    Raised by a slave channel when a chat indicated is not found.

    Can be raised by any method that involves a chat or a message.
    """
    pass


class EFBChannelNotFound(EFBException):
    """
    Raised by the coordinator when the message sent delivers to
    a missing channel.
    """
    pass


class EFBMessageError(EFBException):
    """
    Raised by slave channel for any other error occurred when sending
    a message or a status.

    Can be raised in :meth:`.EFBChannel.send_message` and :meth:`.EFBChannel.send_status`.
    """
    pass


class EFBMessageNotFound(EFBMessageError):
    """
    Raised by a slave channel when a message indicated is not found.

    Can be raised in :meth:`.EFBChannel.send_message` (edited message / target message not found)
    and in :meth:`.EFBChannel.send_status` (message to delete is not found).
    """
    pass


class EFBMessageTypeNotSupported(EFBMessageError):
    """
    Raised by a slave channel when the indicated message type is not supported.

    Can be raised in :meth:`.EFBChannel.send_message`.
    """
    pass


class EFBOperationNotSupported(EFBMessageError):
    """
    Raised by slave channels when a chat operation is not supported.
    E.g.: cannot edit message, cannot delete message.

    Can be raised in :meth:`.EFBChannel.send_message` and :meth:`.EFBChannel.send_status`.
    """
    pass



# coding=utf-8

from .channel import EFBChannel
from .chat import EFBChat
from .constants import ChatType, ChannelType, MsgType, TargetType
from .message import EFBMsg
from .status import EFBStatus
from .middleware import EFBMiddleware
from .__version__ import __version__



# coding=utf-8

from abc import ABC, abstractmethod
from typing import Optional, Dict, List, Set, Callable, IO, TYPE_CHECKING
from .constants import *

if TYPE_CHECKING:
    from .chat import EFBChat
    from .message import EFBMsg
    from .status import EFBStatus

__all__ = ["EFBChannel"]


class EFBChannel(ABC):
    """
    The abstract channel class.

    Attributes:
        channel_name (str):
            A human-friendly name of the channel.
        channel_emoji (str):
            Emoji icon of the channel. Recommended to use a
            visually-length-one emoji that represents
            the channel best.
        channel_type (:obj:`.ChannelType`): Type of the channel.
        supported_message_types (Set[:obj:`.MsgType`]):
            Types of messages that the channel accepts as incoming messages.
        channel_id (str):
            Unique identifier of the channel.
            Convention of IDs is specified in :doc:`/guide/packaging`.
            This ID will be appended with its instance ID when available.
        instance_id (str):
            The instance ID if available.
    """

    channel_name: str = "Empty channel"
    channel_emoji: str = "�"
    channel_id: str = "efb.empty_channel"
    channel_type: ChannelType = None
    instance_id: str = None
    supported_message_types: Set[MsgType] = set()
    __version__: str = 'undefined version'

    def __init__(self, instance_id: str = None):
        """
        Initialize the channel.
        Inherited initializer must call the "super init" method
        at the beginning.

        Args:
            instance_id: Instance ID of the channel.
        """
        self.instance_id = instance_id
        if instance_id:
            self.channel_id += f"#{instance_id}"

    def get_extra_functions(self) -> Dict[str, Callable]:
        """Get a list of additional features

        Returns:
            Dict[str, Callable]: A dict of methods marked as additional features.
            Method can be called with ``get_extra_functions()["methodName"]()``.
        """
        if self.channel_type == ChannelType.Master:
            raise NameError("get_extra_function is not available on master channels.")
        methods = {}
        for mName in dir(self):
            m = getattr(self, mName)
            if getattr(m, "extra_fn", False):
                methods[mName] = m
        return methods

    @abstractmethod
    def send_message(self, msg: 'EFBMsg') -> 'EFBMsg':
        """send_message(msg: EFBMsg) -> EFBMsg

        Send a message to, or edit a sent message in
        the channel.

        Args:
            msg (:obj:`.EFBMsg`): Message object to be processed.

        Returns:
            :obj:`.EFBMsg`:
                The same message object with message ID from the
                recipient.

        Raises:
            :exc:`~.exceptions.EFBChatNotFound`: Raised when a chat
                required is not found.
            :exc:`~.exceptions.EFBMessageTypeNotSupported`: Raised
                when the message type sent is not supported by the
                channel.
            :exc:`~.exceptions.EFBOperationNotSupported`: Raised
                when an message edit request is sent, but not
                supported by the channel.
            :exc:`~.exceptions.EFBMessageNotFound`: Raised when
                an existing message indicated is not found.
                E.g.: The message to be edited, the message referred
                in the :attr:`msg.target <.EFBMsg.target>`
                attribute.
            :exc:`~.exceptions.EFBMessageError`: Raised when other
                error occurred while sending or editing the
                message.
        """
        raise NotImplementedError()

    @abstractmethod
    def poll(self):
        """
        Method to poll for messages. This method is called when
        the framework is initialized. This method should be blocking.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_chats(self) -> List['EFBChat']:
        """get_chats() -> List[EFBChat]

        Return a list of available chats in the channel.

        Returns:
            List[.EFBChat]: a list of available chats in the channel.
        
        Note:
            This is not required by Master Channels
        """
        raise NotImplementedError()

    @abstractmethod
    def get_chat(self, chat_uid: str, member_uid: Optional[str] = None) -> 'EFBChat':
        """get_chat(self, chat_uid: str, member_uid: Optional[str] = None) -> EFBChat

        Get the chat object from a slave channel.
        
        Args:
            chat_uid (str): UID of the chat.
            member_uid (Optional[str]): UID of group member,
                only when the selected chat is a group.

        Returns:
           .EFBChat: The chat found.

        Raises:
            :exc:`~.exceptions.EFBChatNotFound`: Raised when a chat
                required is not found.

        Note:
            This is not required by Master Channels
        """
        raise NotImplementedError()

    @abstractmethod
    def send_status(self, status: 'EFBStatus'):
        """
        Send a status to the channel.

        Args:
            status (:obj:`.EFBStatus`): the status

        Raises:
            :exc:`~.exceptions.EFBChatNotFound`: Raised when a chat
                required is not found.
            :exc:`~.exceptions.EFBMessageNotFound`: Raised when
                an existing message indicated is not found.
                E.g.: The message to be removed.
            :exc:`~.exceptions.EFBOperationNotSupported`: Raised
                when the channel does not support message removal.
            :exc:`~.exceptions.EFBMessageError`: Raised when other
                error occurred while removing the message.

                .. note::
                    Avoid raising exceptions from this method
                    in Master Channels as it would be hard
                    for a Slave Channel to process the
                    exception.

        Note:
            This is not applicable to Slave Channels
        """
        raise NotImplementedError()

    @abstractmethod
    def get_chat_picture(self, chat: 'EFBChat') -> IO[bytes]:
        """get_chat_picture(chat: EFBChat) -> IO[bytes]

        Get the profile picture of a chat. Profile picture is
        also referred as profile photo, avatar, "head image"
        sometimes.
        
        Args:
            chat (.EFBChat): Chat to get picture from.

        Returns:
            IO[bytes]: Opened temporary file object.
            The file object must have appropriate extension name
            that matches to the format of picture sent,
            and seek to position 0.

            It can be deleted when closed if not required otherwise.

        Raises:
            :exc:`~.exceptions.EFBChatNotFound`: Raised when a chat
                required is not found.
            :exc:`~.exceptions.EFBOperationNotSupported`: Raised
                when the chat does not offer a profile picture.
            
        Examples:
            .. code:: Python
            
                if chat.channel_uid != self.channel_uid:
                    raise EFBChannelNotFound()
                file = tempfile.NamedTemporaryFile(suffix=".png")
                response = requests.post("https://api.example.com/get_profile_picture/png",
                                         data={"uid": chat.chat_uid})
                if response.status_code == 404:
                    raise EFBChatNotFound()
                file.write(response.content)
                file.seek(0)
                return file
                
        Note:
            This is not required by Master Channels
        """
        raise NotImplementedError()

    def stop_polling(self):
        """
        When EFB framework is asked to stop gracefully,
        this method is called to each channel object to
        stop all processes in the channel, save all
        status if necessary, and terminate polling.

        When the channel is ready to stop, the polling
        function must stop blocking. EFB framework will
        quit completely when all polling threads end.
        """
        raise NotImplementedError()



# coding=utf-8

from abc import ABC
from typing import Optional

from .message import EFBMsg
from .status import EFBStatus

__all__ = ['EFBMiddleware']


class EFBMiddleware(ABC):
    """
    Middleware class.

    Attributes:
        middleware_id (str):
            Unique ID of the middleware
            Convention of IDs is specified in :doc:`/guide/packaging`.
            This ID will be appended with its instance ID when available.
        middleware_name (str): Human-readable name of the middleware.
        instance_id (str):
            The instance ID if available.
    """
    middleware_id: str = "efb.empty_middleware"
    middleware_name: str = "Empty Middleware"
    instance_id: str = None
    __version__: str = 'undefined version'

    def __init__(self, instance_id: str = None):
        """
        Initialize the middleware.
        Inherited initializer must call the "super init" method
        at the beginning.

        Args:
            instance_id: Instance ID of the channel.
        """
        self.instance_id = instance_id
        if instance_id:
            self.middleware_id += f"#{instance_id}"

    def process_message(self, message: EFBMsg) -> Optional[EFBMsg]:
        """
        Process a message with middleware

        Args:
            message (:obj:`.EFBMsg`): Message object to process

        Returns:
            Optional[:obj:`.EFBMsg`]: Processed message or None if discarded.
        """
        return message

    def process_status(self, status: EFBStatus) -> Optional[EFBStatus]:
        """
        Process a status update with middleware

        Args:
            status (:obj:`.EFBStatus`): Message object to process

        Returns:
            Optional[:obj:`.EFBStatus`]: Processed status or None if discarded.
        """
        return status



# coding=utf-8

import os

import sys
import yaml
from typing import Dict, Any
from . import utils
from .channel import EFBChannel
from .middleware import EFBMiddleware
from .constants import ChannelType

__all__ = ["load_config"]


def load_config() -> Dict[str, Any]:
    # Include custom channels
    custom_channel_path = utils.get_custom_modules_path()
    if custom_channel_path not in sys.path:
        sys.path.insert(0, custom_channel_path)

    conf_path = utils.get_config_path()
    if not os.path.exists(conf_path):
        raise FileNotFoundError("Config File does not exist. (%s)" % conf_path)
    with open(conf_path) as f:
        data = yaml.load(f.read())

        # Verify configuration

        # - Master channel
        if not isinstance(data.get("master_channel", None), str):
            raise ValueError("Master Channel path must be a string.")
        channel = utils.locate_module(data['master_channel'], 'master')
        if not channel:
            raise ValueError("\"%s\" is not found." % data['master_channel'])
        if not issubclass(channel, EFBChannel):
            raise ValueError("\"%s\" is not a channel." % data['master_channel'])
        if not channel.channel_type == ChannelType.Master:
            raise ValueError("\"%s\" is not a master channel." % data['master_channel'])

        # - Slave channels
        if not isinstance(data.get("slave_channels", None), list):
            raise ValueError("Slave Channel paths must be a list.")
        for i in data['slave_channels']:
            channel = utils.locate_module(i, 'slave')
            if not channel:
                raise ValueError("\"%s\" is not found." % i)
            if not issubclass(channel, EFBChannel):
                raise ValueError("\"%s\" is not a channel." % i)
            if not channel.channel_type == ChannelType.Slave:
                raise ValueError("\"%s\" is not a slave channel." % i)

        # - Middlewares
        if data.get("middlewares", None) is not None:
            if not isinstance(data.get("middlewares"), list):
                raise ValueError("Middleware paths must be a list")
            for i in data['middlewares']:
                middleware = utils.locate_module(i, 'middleware')
                if not middleware:
                    raise ValueError("\"%s\" is not found." % i)
                if not issubclass(middleware, EFBMiddleware):
                    raise ValueError("\"%s\" is not a middleware." % i)
        else:
            data['middlewares'] = list()
    return data



# coding=utf-8

from abc import ABC, abstractmethod
from typing import IO, Dict, Optional, List, Any, Tuple

from .constants import *
from .chat import EFBChat
from .channel import EFBChannel


class EFBMsg:
    """A message.

    Attributes:
        attributes (Optional[:obj:`.EFBMsgAttribute`]):
            Attributes used for a specific message type.
            Only specific message type requires this attribute. Defaulted to
            ``None``.

            - Link: :obj:`.EFBMsgLinkAttribute`
            - Location: :obj:`.EFBMsgLocationAttribute`
            - Status: Typing/Sending files/etc.: :obj:`.EFBMsgStatusAttribute`

            .. Note::
                Do NOT use object of the abstract class
                :class:`.EFBMsgAttribute` for
                ``attributes``, but object of specific class instead.


        author (:obj:`.EFBChat`): Author of this message.
        chat (:obj:`.EFBChat`): Sender of the message.
        commands (Optional[:obj:`EFBMsgCommands`]): Commands attached to the message
        deliver_to (:obj:`.EFBChannel`): The channel that the message is to be delivered to
        edit (bool): Flag this up if the message is edited.
        file (IO[bytes]): File object to multimedia file, type "ra". ``None`` if N/A.
            recommended to use ``NamedTemporaryFile`` object, the file can be
            deleted when closed, if not used otherwise.
            All file object must be rewind back to 0 (``file.seek(0)``) before sending.
        filename (str): File name of the multimedia file. ``None`` if N/A
        is_system (bool): Mark as true if this message is a system message.
        mime (str): MIME type of the file. ``None`` if N/A
        path (str): Local path of multimedia file. ``None`` if N/A
        substitutions (Optional[:obj:`EFBMsgSubstitutions`]):
            Substitutions of messages, usually used when
            the some parts of the text of the message
            refers to another user or chat.
        target (Optional[:obj:`EFBMsg`]):
            Target message (usually for messages that "replies to" 
            another message).

            .. note::

                This message may be a "minimum message", with only required fields:

                - :attr:`.EFBMsg.chat`
                - :attr:`.EFBMsg.author`
                - :attr:`.EFBMsg.text`
                - :attr:`.EFBMsg.type`
                - :attr:`.EFBMsg.uid`

        text (str): text of the message
        type (:obj:`.MsgType`): Type of message
        uid (str): Unique ID of message.
            Usually stores the message ID from slave channel.
            This ID must be unique among all chats in the same channel.

            .. Note::
                Some channels may not support message editing.
                Some channels may issue a new uid for edited message.

        vendor_specific (Dict[str, Any]):
            A series of vendor specific attributes attached. This can be
            used by any other channels or middlewares that is compatible
            with such information. Note that no guarantee is provided
            for information in this section.

    """
    def __init__(self):
        self.attributes: Optional[EFBMsgAttribute] = None
        self.author: EFBChat = None
        self.chat: EFBChat = None
        self.commands: Optional[EFBMsgCommands] = None
        self.deliver_to: EFBChannel = None
        self.edit: bool = False
        self.file: Optional[IO[bytes]] = None
        self.filename: Optional[str] = None
        self.is_system: bool = False
        self.mime: Optional[str] = None
        self.path: Optional[str] = None
        self.substitutions: Optional[EFBMsgSubstitutions] = None
        self.target: Optional[EFBMsg] = None
        self.text: str = ""
        self.type: MsgType = None
        self.uid: Optional[str] = None
        self.vendor_specific: Dict[str, Any] = dict()

    def __str__(self):
        return "<EFBMsg, {msg.author}@{msg.chat} [{msg.type.name}]: {msg.text}; {msg.uid}>".format(msg=self)

    def __repr__(self):
        return "<EFBMsg, {msg.author}@{msg.chat} [{msg.type.name}]: " \
               "{msg.text}; " \
               "Attributes: {msg.attributes}; " \
               "Delivering to: {msg.deliver_to}; " \
               "Edited: {msg.edit}; " \
               "System message: {msg.is_system}; " \
               "Substitutions: {msg.substitutions}; " \
               "Target messages: {msg.target}; " \
               "UID: {msg.uid};" \
               "File: {msg.file} ({msg.filename} @ {msg.path}), {msg.mime}; " \
               "Vendor: {msg.vendor_specific}>".format(msg=self)

    def verify(self):
        """
        Verify the validity of message.
        """
        if self.author is None or not isinstance(self.author, EFBChat):
            raise ValueError("Author is not valid.")
        else:
            self.author.verify()
        if self.author is None or not isinstance(self.author, EFBChat):
            raise ValueError("Chat is not valid.")
        else:
            self.author.verify()
        if self.type is None or not isinstance(self.type, MsgType):
            raise ValueError("Type is not valid.")
        if self.deliver_to is None or not isinstance(self.deliver_to, EFBChannel):
            raise ValueError("Deliver_to is not valid.")
        if self.type in (MsgType.Audio, MsgType.File, MsgType.Image, MsgType.Sticker, MsgType.Video):
            if self.file is None or not hasattr(self.file, "read") or not hasattr(self.file, "close"):
                raise ValueError("File is not valid.")
            if self.mime is None or not self.mime:
                raise ValueError("MIME is not valid.")
            if self.path is None or not self.path:
                raise ValueError("Path is not valid.")
        if self.type == MsgType.Location and (self.attributes is None or not isinstance(self.attributes, EFBMsgLocationAttribute)):
            raise ValueError("Attribute of location message is invalid.")
        if self.type == MsgType.Link and (self.attributes is None or not isinstance(self.attributes, EFBMsgLinkAttribute)):
            raise ValueError("Attribute of link message is invalid.")
        if self.type == MsgType.Status and (self.attributes is None or not isinstance(self.attributes, EFBMsgStatusAttribute)):
            raise ValueError("Attribute of status message is invalid.")

        if self.attributes:
            self.attributes.verify()

        if self.commands:
            self.commands.verify()

        if self.substitutions:
            self.substitutions.verify()


class EFBMsgAttribute(ABC):
    """Abstract class of a message attribute."""
    @abstractmethod
    def __init__(self):
        raise NotImplementedError("Do not use the abstract class EFBMsgAttribute")

    @abstractmethod
    def verify(self):
        raise NotImplementedError()


class EFBMsgLinkAttribute(EFBMsgAttribute):
    """
    EFB link message attribute.

    Attributes:
        title (str): Title of the link.
        description (str, optional): Description of the link.
        image (str, optional): Image/thumbnail URL of the link.
        url (str): URL of the link.
    """
    title: str = ""
    description: Optional[str] = None
    image: Optional[str] = None
    url: str = ""

    def __init__(self, title: str = None, description: Optional[str] = None,
                 image: Optional[str] = None, url: str = None):
        """
        Args:
            title (str): Title of the link.
            description (str, optional): Description of the link.
            image (str, optional): Image/thumbnail URL of the link.
            url (str): URL of the link.
        """
        if title is None or url is None:
            raise ValueError("Title and URL is required.")
        self.title = title
        self.description = description
        self.image = image
        self.url = url

    def __str__(self):
        return "<EFBMsgLinkAttribute, {attr.title}: {attr.description} " \
               "({attr.image}) @ {attr.url}>".format(attr=self)

    def verify(self):
        if self.url is None:
            raise ValueError("URL does not exist")
        if self.title is None:
            raise ValueError("Title does not exist")


class EFBMsgLocationAttribute(EFBMsgAttribute):
    """
    EFB location message attribute.

    Attributes:
        latitude (float): Latitude of the location.
        longitude (float): Longitude of the location.
    """
    latitude: float = 0
    longitude: float = 0

    def __init__(self, latitude: float, longitude: float):
        """
        Args:
            latitude (float): Latitude of the location.
            longitude (float): Longitude of the location.
        """
        self.latitude = latitude
        self.longitude = longitude

    def __str__(self):
        return "<EFBMsgLocationAttribute: {attr.latitude}, {attr.longitude}>".format(attr=self)

    def verify(self):
        if self.latitude is None or isinstance(self.latitude, float):
            raise ValueError("Latitude is invalid.")
        if self.latitude is None or isinstance(self.latitude, float):
            raise ValueError("Longitude is invalid.")


class EFBMsgCommand:
    """
    EFB message command.

    Attributes:
        name (str): Human-friendly name of the command.
        callable_name (str): Callable name of the command.
        args (List[Any]): Arguments passed to the function.
        kwargs (Dict[str, Any]): Keyword arguments passed to the function.
    """
    name: str = ""
    callable_name: str = ""
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}

    def __init__(self, name: str, callable_name: str, args: List[Any]=None, kwargs: Optional[Dict[str, Any]]=None):
        """
        Args:
            name (str): Human-friendly name of the command.
            callable_name (str): Callable name of the command.
            args (Optional[List[Any]]): Arguments passed to the function. Defaulted to empty list;
            kwargs (Optional[Dict[str, Any]]): Keyword arguments passed to the function.
                Defaulted to empty dict.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if not isinstance(name, str):
            raise TypeError("name must be a string.")
        if not isinstance(callable_name, str):
            raise TypeError("callable must be a string.")
        if not isinstance(args, list):
            raise TypeError("args must be a list.")
        if not isinstance(kwargs, dict):
            raise TypeError("kwargs must be a dict.")
        self.name = name
        self.callable_name = callable_name
        self.args = args.copy()
        self.kwargs = kwargs.copy()

    def __str__(self):
        return "<EFBMsgCommand: {name}, {callable_name}({params})>".format(
            name=self.name,
            callable_name=self.callable_name,
            params=", ".join(self.args + ["%r=%r" % i for i in self.kwargs.items()])
        )

    def verify(self):
        if not self.name:
            raise ValueError("Name does not exist.")
        if not self.callable_name:
            raise ValueError("Callable does not exist.")


class EFBMsgCommands:
    """
    EFB message commands.
    Message commands allow user to take action to
    a specific message, including vote, add friends, etc.

    Attributes:
        commands (list of :obj:`EFBMsgCommand`): Commands for the message.
    """

    commands: List[EFBMsgCommand] = []

    def __init__(self, commands: List[EFBMsgCommand]):
        """
        Args:
            commands (list of :obj:`EFBMsgCommand`): Commands for the message.
        """
        if not (isinstance(commands, list) and len(commands) > 0 and all(
                isinstance(i, EFBMsgCommand) for i in commands)):
            raise ValueError("There must be one or more commands, "
                             "and all of them must be in type EFBMsgCommand.")
        self.commands = commands.copy()

    def __str__(self):
        return str(self.commands)

    def verify(self):
        for i in self.commands:
            i.verify()


class EFBMsgStatusAttribute(EFBMsgAttribute):
    """
    EFB Message status attribute.
    Message with type ``Status`` notifies the other end to update a chat-specific
    status, such as typing, send files, etc.

    Attributes:
        status_type: Type of status, possible values are defined in the
            ``EFBMsgStatusAttribute``.
        timeout (Optional[int]):
                Number of milliseconds for this status to expire.
                Default to 5 seconds.
        Types: List of status types supported
    """
    class Types(Enum):
        """
        Attributes:
            TYPING:
                Used in :attr:`~.ehforwarderbot.message.EFBMsgStatusAttribute.status_type`,
                represent the status of typing.
            UPLOADING_FILE:
                Used in :attr:`~.ehforwarderbot.message.EFBMsgStatusAttribute.status_type`,
                represent the status of uploading file.
            UPLOADING_IMAGE:
                Used in :attr:`~.ehforwarderbot.message.EFBMsgStatusAttribute.status_type`,
                represent the status of uploading image.
            UPLOADING_AUDIO:
                Used in :attr:`~.ehforwarderbot.message.EFBMsgStatusAttribute.status_type`,
                represent the status of uploading audio.
            UPLOADING_VIDEO:
                Used in :attr:`~.ehforwarderbot.message.EFBMsgStatusAttribute.status_type`,
                represent the status of uploading video.
        """
        TYPING = "TYPING"
        UPLOADING_FILE = "UPLOADING_FILE"
        UPLOADING_IMAGE = "UPLOADING_IMAGE"
        UPLOADING_AUDIO = "UPLOADING_AUDIO"
        UPLOADING_VIDEO = "UPLOADING_VIDEO"

    def __init__(self, status_type: Types, timeout: Optional[int] = 5000):
        """
        Args:
            status_type: Type of status.
            timeout (Optional[int]):
                Number of milliseconds for this status to expire.
                Default to 5 seconds.
        """
        self.status_type: self.Types = status_type
        self.timeout: int = timeout

    def __str__(self):
        return "<EFBMsgStatusAttribute: {attr.status_type} @ {attr.timeout}ms>".format(attr=self)

    def verify(self):
        if self.status_type is None or not isinstance(self.status_type, self.Types):
            raise ValueError("Status type is invalid.")
        if not isinstance(self.timeout, int):
            raise ValueError("Timeout is invalid.")


class EFBMsgSubstitutions(dict):
    """
    EFB message substitutions.

    This is for the case when user "@-referred" a list of users in the message.
    Substitutions here is a dict of correspondence between
    the string used to refer to the user in the message
    and a user object.

    Dictionary of text substitutions targeting to a user or member.

    The key of the dictionary is a tuple of two :obj:`int`\ s, where first
    of it is the starting position in the string, and the second is the
    ending position defined similar to Python's substring. A tuple of
    ``(3, 15)`` corresponds to ``msg.text[3:15]``.
    The value of the tuple ``(a, b)`` must lie within ``a ∈ [0, l)``,
    ``b ∈ (a, l]``, where ``l`` is the length of the message text.

    Value of the dict may be any user of the chat, or a member of a
    group. Notice that the :obj:`EFBChat` object here must NOT be a
    group.

    Type:
        Dict[Tuple[int, int], :obj:`.EFBChat`]
    """

    def __init__(self, substitutions: Dict[Tuple[int, int], EFBChat]):
        if not isinstance(substitutions, dict):
            raise TypeError("Substitutions must be a dict.")
        for i in substitutions:
            if not isinstance(i, tuple) or not len(i) == 2 or not isinstance(i[0], int) or not isinstance(i[1], int) \
                    or not i[0] < i[1]:
                raise TypeError("Substitution %s's index must be a tuple of 2 integers where the first one is less"
                                "than the second one." % i)
            if not isinstance(substitutions[i], EFBChat):
                raise TypeError("Substitution %s is not a chat object." % i)
            if substitutions[i].is_chat and \
                            substitutions[i].chat_type == ChatType.Group:
                raise ValueError("Substitution %s is a group." % i)
        super().__init__(substitutions)

    def verify(self):
        ranges = sorted(self.keys())
        if ranges and (ranges[0][0] < 0 or ranges[0][1] < ranges[0][0]):
            raise ValueError("Index %s is invalid." % ranges[0])
        for i in range(1, len(ranges)):
            if ranges[i][0] < 0 or ranges[i][1] < ranges[i][0]:
                raise ValueError("Index %s is invalid." % ranges[i])
            if ranges[i][0] < ranges[i - 1][1]:
                raise ValueError("Index %s overlaps with %s." % (ranges[i], ranges[i - 1]))
        for i in self.values():
            i.verify()